local M = {}

local scenario = require("scenario/scenarios")
local helper = require("scenario/scenariohelper")

local playerName = "rr_player0"
local policeNames = {"rr_police0", "rr_police1", "rr_police_sheriff0"}
local rampName = "rr_ramp"
local waypointNames = {"waypoint0", "waypoint1", "waypoint2", "waypoint3", "waypoint4", "waypoint5"}
local finishName = "waypoint5"
local messages = {fail = "Busted! Get in the handcuffs.", win = "Phew, you evaded them! Enjoy the gold!", quickWin = "Phew, you evaded them! That was impressively fast.", skillWin = "Phew, you evaded them! Looks like you didn't need any help from ramps!"}

local speedLimit = 30 --Speed limit of lawful drivers
local zonesTotal = 20 --Number of spawn zones in the prefab
local policeAggro = 1.2 --The aggression level of the police
local nearDist = 60 --The distance that you can get arrested at
local countMax = 5 --Arrest countdown timer start value
local mainTimer = 0
local zoneCheck = 0
local zoneStart = 0
local hornHold = 0
local delay = 0
local funMode = 0
local activator = nil
local allowRoadblock = false
local setRoadblock = false
local allowRamp = false
local setRamp = false
local stockRamp = 0
local escape = nil
local slowMotion = false

local Police = {}
local player = {}
local police = {}
local ramp = {}
local zones = {}

function Police:new(obj, damage, special, state, zone, timer)
	local self = setmetatable({}, Police)
	self.__index = self
	self.obj = obj or nil --The vehicle itself
	self.damage = damage or nil --Vehicle overall damage
	self.special = special or nil --Determines if the vehicle has special properties
	self.state = state or nil --Behavior states; none, cruise, ready, wait, block
	self.zone = zone or nil --Zone that the vehicle is currently in
	self.timer = timer or nil --Timer, used for delays
	return self
end

local function arrestPlayer(vehicle) --Arrests the player with a countdown
	local countdown = 0
	vehicle.timer = vehicle.timer + 0.25
	if vehicle.timer % 1 == 0 then
		local countdown = countMax - math.modf(vehicle.timer)
		if countdown > 0 then
			helper.flashUiMessage((tostring(countdown).."..."), 0.5, true)
		end
	end
	if vehicle.timer > countMax then
		scenario.finish({failed = messages.fail})
	end
end

local function preSpawn(vehicle) --Spawns the vehicle under the spawnzone, preparing it for position and rotation
	local zoneIn = scenetree.findObject(tostring("spawnzone"..(player.zone + 2 - funMode)))
	if vehicle.obj.name == rampName then
		zoneIn = scenetree.findObject(police[1].zone)
	end
	local zonePos = zoneIn:getPosition()
	local zoneRot = zoneIn:getRotation()
	vehicle.obj:setPositionRotation(zonePos.x, zonePos.y, zonePos.z - 5, 0, 0, zoneRot.z, zoneRot.w)
	log("Z", "gull", vehicle.obj.name.." teleported to "..zoneIn.name)
	vehicle.state = "wait"
	vehicle.zone = zoneIn.name
	vehicle.timer = 0
end

local function postSpawn(vehicle) --Spawns the vehicle into the roadblock formation
	local ra = 0
	local vert = 0
	local posOffset = nil
	local rotOffset = nil
	local zonePos = scenetree.findObject(vehicle.zone):getPosition()
	for i, v in ipairs(police) do
		if vehicle.obj.name == police[i].obj.name then
			ra = (i * 2) - 3
			vert = 0.5
			posOffset = vec3(vehicle.obj:getPosition()) - vec3(vehicle.obj:getDirectionVector()):rotated(quat(0, 0, 0.7071 * ra, 0.7071)) * 3.333
			rotOffset = quatFromDir(vec3(vehicle.obj:getDirectionVector()):rotated(quat(0, 0, 0.7071 * ra, 0.7071)), vec3(0,0,1))
			if vehicle.damage >= 1000 then
				vehicle.obj:queueLuaCommand("vehicleResetted(false)")
			end
		end
	end
	if vehicle.obj.name == rampName then
		vehicle.fix = 0
		vert = 2
		posOffset = vec3(vehicle.obj:getPosition()) - vec3(vehicle.obj:getDirectionVector()):rotated(quat(0, 0, -0.1, 0.7071)) * 9
		rotOffset = quatFromDir(vec3(vehicle.obj:getDirectionVector()):rotated(quat(0, 0, 1, 0)), vec3(0,0,1))
	end
	vehicle.obj:setPositionRotation(posOffset.x, posOffset.y, zonePos.z + vert, 0, 0, rotOffset.z, rotOffset.w)
	vehicle.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
	vehicle.state = "block"
	vehicle.timer = 0
end

local function fixRamp(vehicle)
	local tempPos = vehicle.obj:getPosition()
	local tempRot = vehicle.obj:getRotation()
	vehicle.obj:setPositionRotation(tempPos.x, tempPos.y, tempPos.z - 0.08, tempRot.x, tempRot.y, tempRot.z, tempRot.w)
end

local function onScenarioUIReady()
	player = {obj = scenetree.findObject(playerName), pos = nil, vel = nil, dist = nil, zone = 0, timer = 0}
	ramp = {obj = scenetree.findObject(rampName), pos = nil, fix = 0, state = nil, zone = 0, timer = 0}
	for i, v in ipairs(policeNames) do
		police[i] = Police:new(scenetree.findObject(policeNames[i]), 0, "normal", "cruise", nil, 0)
	end
	police[3].special = "special"
	for  i = 1, zonesTotal, 1 do
		zones[i] = "spawnzone"..(i - 1)
	end
	activator = scenetree.findObject("activator")
	allowRoadblock = false
	setRoadblock = false
	allowRamp = false
	setRamp = false
	stockRamp = 3
	mainTimer = 0
	zoneCheck = 0
	zoneStart = math.random(0, 2)
	hornHold = 0
	funMode = 0
	escape = nil
	slowMotion = false
end

local function onRaceStart()
	for i, v in ipairs(police) do
		if police[i].special == "normal" then
			police[i].obj:queueLuaCommand("ai.setMode('random')")
			police[i].obj:queueLuaCommand("ai.driveInLane('on')")
			police[i].obj:queueLuaCommand("ai.setSpeed("..tostring(speedLimit)..")")
			police[i].obj:queueLuaCommand("ai.setSpeedMode('limit')")
		else
			police[i].obj:queueLuaCommand("ai.setMode('chase')")
			police[i].obj:queueLuaCommand("ai.setAggression("..tostring(policeAggro)..")")
			police[i].obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
		end
	end
end

local function onRaceTick(raceTickTime)
	mainTimer = mainTimer + raceTickTime
	player.pos = player.obj:getPosition()
	player.vel = map.objects[map.objectNames[player.obj.name]].vel:length()
	player.markerDist = (player.pos - activator:getPosition()):len()
	player.obj:queueLuaCommand("obj:queueGameEngineLua('zHorn = ('..electrics.values.horn..')')") --Creates a global variable for the horn state; 0 = off and 1 = on
	if zHorn == 1 then
		hornHold = hornHold + raceTickTime
		if mainTimer == 1 then
			helper.flashUiMessage("Ramps not ready yet!", 2, false)
		end
		if mainTimer <= 10 and hornHold == 5 then
			helper.flashUiMessage("Fun mode enabled!", 2, false)
			funMode = 1
		end
	end
	if player.markerDist <= 24 and allowRoadblock == false then
		helper.flashUiMessage("Warning, roadblocks ahead!", 3, false)
		for i, v in ipairs(police) do
			if police[i].special == "normal" then
				police[i].obj:queueLuaCommand("ai.setMode('stopping')")
			end
		end
		allowRoadblock = true
		setRoadblock = true
		allowRamp = true
		ramp.state = "normal"
	end
	if allowRoadblock == true and zoneCheck > zoneStart then
		player.zone = zoneCheck
	else
		player.zone = zoneStart
	end
	if (player.pos - ramp.obj:getPosition()):len() <= nearDist / 2 and ramp.fix == 0 then
		fixRamp(ramp)
		ramp.fix = 1
	end
	if zHorn == 1 and allowRamp == true and setRamp == false then
		if stockRamp <= 0 then
			if allowRamp == true then
				helper.flashUiMessage("You used up all of your ramp calls!", 2, false)
				allowRamp = false
			end
		end
		if stockRamp > 0 then
			helper.flashUiMessage("Ramp deployed!", 2, false)
			setRamp = true
			if funMode == 0 then
				stockRamp = stockRamp - 1
			end
		end
	end
	if ramp.state == "wait" then
		if ramp.timer == 0.5 then
			postSpawn(ramp)
		end
		ramp.timer = ramp.timer + raceTickTime
	end
	if setRoadblock == true and setRamp == true and ramp.state == "normal" then
		preSpawn(ramp)
	end
	for i, v in ipairs(police) do
		police[i].pos = police[i].obj:getPosition()
		police[i].vel = map.objects[map.objectNames[police[i].obj.name]].vel:length()
		police[i].damage = map.objects[map.objectNames[police[i].obj.name]].damage
		if police[i].special == "normal" then
			if mainTimer * 2 == i then
				police[i].obj:queueLuaCommand("electrics.set_lightbar_signal(2)") --Delays the AI police sirens by a bit; sounds nicer!
			end
			if allowRoadblock == false and police[i].state == "cruise" then
				police[i].obj:queueLuaCommand("ai.setSpeed("..tostring(math.min(speedLimit, player.vel * policeAggro * policeAggro))..")")
			end
			if police[i].state == "wait" then
				if police[i].timer == 0.5 then
					postSpawn(police[i])
				end
				police[i].timer = police[i].timer + raceTickTime
			end
			if setRoadblock == true and police[i].state == "cruise" then
				police[i].state = "ready"
				police[i].timer = 0
			end
			if delay > 0 and police[i].state == "ready" then
				if police[i].timer >= 0.5 and police[i].obj.name == police[delay].obj.name then
					preSpawn(police[i])
					delay = -4
				end
				police[i].timer = police[i].timer + raceTickTime
			end
		end
	end
	delay = delay + 1
	if delay > #police then
		delay = 1
	end
	player.dist = math.min((player.pos - police[1].pos):len(), (player.pos - police[2].pos):len(), (player.pos - police[3].pos):len())
	if mainTimer >= 3 and player.dist <= nearDist and player.vel <= 3 then
		arrestPlayer(player)
	else
		player.timer = 0
	end
	if commands.isFreeCamera(0) then
		if slowMotion == false then
			bullettime.set(0.125)
			slowMotion = true
		end
	else
		if slowMotion == true then
			bullettime.set(1)
			slowMotion = false
		end
	end
end

local function onRaceResult()
	if escape == true then
		if mainTimer <= 120 then
			scenario.finish({msg = messages.quickWin})
		elseif stockRamp == 3 then
			scenario.finish({msg = messages.skillWin})
		else
			scenario.finish({msg = messages.win})
		end
	end
end

local function onRaceWaypointReached(data)
	if data.waypointName == finishName then
		if data.vehicleName == player.obj.name and escape == nil then
			escape = true
		end
	end
end

local function onBeamNGTrigger(data)
	if data.event == "enter" then
		if data.subjectName == player.obj.name and string.byte(tostring(data.triggerName)) == 115 then
			for i, v in ipairs(zones) do
				if data.triggerName == zones[i] then
					zoneCheck = i
				end
			end
			if zoneCheck > zonesTotal - 3 then
				zoneCheck = 0
			end
		end
	end
	if data.event == "exit" then
		if data.subjectName == player.obj.name and string.byte(tostring(data.triggerName)) == 115 then
			if setRoadblock == true and data.triggerName == police[1].zone then
				for i, v in ipairs(police) do
					police[i].state = "cruise"
				end
				ramp.state = "normal"
				setRamp = false
			end
		end
	end
end

M.onScenarioUIReady = onScenarioUIReady
M.onRaceStart = onRaceStart
M.onRaceTick = onRaceTick
M.onRaceResult = onRaceResult
M.onRaceWaypointReached = onRaceWaypointReached
M.onBeamNGTrigger = onBeamNGTrigger

return M